#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import re

import pytest
import time
from pybis import DataSet
from pybis import Openbis


def test_create_delete_space(openbis_instance):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    space_name = "test_space_" + timestamp
    space = openbis_instance.new_space(code=space_name)
    space.save()
    space_exists = openbis_instance.get_space(code=space_name)
    assert space_exists is not None

    space.delete("test on {}".format(timestamp))

    with pytest.raises(ValueError):
        space_not_exists = openbis_instance.get_space(code=space_name, use_cache=False)



#   Copyright ETH 2025 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from pandas.testing import assert_frame_equal
from pandas import DataFrame
import pytest
import base64
from pybis.spreadsheet import Spreadsheet
from pybis.property_reformatter import PropertyReformatter


# base64 encoded Spreadsheet with special characters encoded with latin-1 encoding
SPREADSHEET_LATIN1 = '<DATA> </DATA>'

def test_create_simple_spreadsheet():

    spreadsheet = Spreadsheet(5, 5)
    for i in range(5):
        spreadsheet.data[i][i] = i+1
        spreadsheet.values[i][i] = i - 1

    assert spreadsheet.width == [50,50,50,50,50]
    assert len(spreadsheet.style) == 25
    assert spreadsheet.headers == ['A', 'B', 'C', 'D', 'E']
    for i in range(5):
        assert spreadsheet[spreadsheet.headers[i], i+1].formula == i+1
        assert spreadsheet[spreadsheet.headers[i], i+1].value == i-1

def test_add_row():
    spreadsheet = Spreadsheet(3, 3)
    for i in range(3):
        for j in range(3):
            spreadsheet.data[i][j] = (i+1)*(j+1)
            spreadsheet.values[i][j] = (i+1)*(j+1)

    assert len(spreadsheet.data) == 3
    assert len(spreadsheet.values) == 3
    assert len(spreadsheet.style) == 9

    spreadsheet.add_row()

    assert len(spreadsheet.data) == 4
    assert len(spreadsheet.values) == 4
    assert len(spreadsheet.style) == 12

    for i in range(3):
        for j in range(3):
            assert spreadsheet[i+1, j+1].formula == (i+1)*(j+1)
            assert spreadsheet[i+1, j+1].value == (i+1)*(j+1)

def test_delete_row_first():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '2'] = i

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    spreadsheet.delete_row(1)

    assert len(spreadsheet.data) == 4
    assert len(spreadsheet.values) == 4
    assert len(spreadsheet.style) == 20

    for i in range(5):
        for j in range(4):
            assert spreadsheet[i + 1, j+1].formula == (i + 1) * (j+2)
            assert spreadsheet[i + 1, j+1].value == (i + 1) * (j+2)

    for i in range(5):
        assert spreadsheet[columns[i], 1].style == columns[i]


def test_delete_row_last():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '4'] = i

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    spreadsheet.delete_row(5)

    assert len(spreadsheet.data) == 4
    assert len(spreadsheet.values) == 4
    assert len(spreadsheet.style) == 20

    for i in range(5):
        for j in range(4):
            assert spreadsheet[i + 1, j + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[i + 1, j + 1].value == (i + 1) * (j + 1)

    for i in range(5):
        assert spreadsheet[columns[i], 4].style == columns[i]


def test_delete_row_middle():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '4'] = i
        spreadsheet.style[i + '2'] = i

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    spreadsheet.delete_row(3)

    assert len(spreadsheet.data) == 4
    assert len(spreadsheet.values) == 4
    assert len(spreadsheet.style) == 20

    for i in range(5):
        for j in range(2):
            assert spreadsheet[i + 1, j + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[i + 1, j + 1].value == (i + 1) * (j + 1)
        for j in range(3,5,1):
            assert spreadsheet[i + 1, j].formula == (i + 1) * (j + 1)
            assert spreadsheet[i + 1, j].value == (i + 1) * (j + 1)

    for i in range(5):
        assert spreadsheet[columns[i], 2].style == columns[i]
        assert spreadsheet[columns[i], 3].style == columns[i]

def test_add_column_no_name():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '4'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[4] = 123

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    spreadsheet.add_column()

    assert len(spreadsheet.style) == 30
    assert len(spreadsheet.headers) == 6
    assert len(spreadsheet.width) == 6
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    for i in range(5):
        assert len(spreadsheet.data[i]) == 6
        assert len(spreadsheet.values[i]) == 6

    assert spreadsheet.headers == ['A', 'B', 'C', 'D', 'E', 'F']
    assert spreadsheet.width == [50, 50, 50, 50, 123, 50]

    for i in range(5):
        for j in range(5):
            assert spreadsheet[i + 1, j + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[i + 1, j + 1].value == (i + 1) * (j + 1)
        assert spreadsheet[6, i+1].formula == ''
        assert spreadsheet[6, i+1].value == ''
        assert spreadsheet[6, i+1].style == 'text-align: center;'

def test_add_column_with_name():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '4'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[4] = 123

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    spreadsheet.add_column('OPENBIS')

    assert len(spreadsheet.style) == 30
    assert len(spreadsheet.headers) == 6
    assert len(spreadsheet.width) == 6
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5

    for i in range(5):
        assert len(spreadsheet.data[i]) == 6
        assert len(spreadsheet.values[i]) == 6

    assert spreadsheet.headers == ['A', 'B', 'C', 'D', 'E', 'OPENBIS']
    assert spreadsheet.width == [50, 50, 50, 50, 123, 50]

    for i in columns:
        assert spreadsheet.style[i + '4'] == i
        assert spreadsheet.style[i + '2'] == i

    for i in range(5):
        for j in range(5):
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 1)
        assert spreadsheet['OPENBIS', i+1].formula == ''
        assert spreadsheet['OPENBIS', i+1].value == ''
        assert spreadsheet['OPENBIS', i+1].style == 'text-align: center;'

def test_delete_column_first_digit():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[1] = 123

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column(1)

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [123, 50, 50, 50]
    assert spreadsheet.headers == ['B', 'C', 'D', 'E']

    for j in range(4):
        for i in range(5):
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 2)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 2)

    for i in ['B', 'C', 'D', 'E']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i


def test_delete_column_first_column_name():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[1] = 123

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column('A')

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [123, 50, 50, 50]
    assert spreadsheet.headers == ['B', 'C', 'D', 'E']

    for i in range(5):
        for j in range(4):
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 2)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 2)

    for i in ['B', 'C', 'D', 'E']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i

def test_delete_column_last_digit():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[4] = 123
    spreadsheet.width[3] = 321

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column(5)

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [50, 50, 50, 321]
    assert spreadsheet.headers == ['A', 'B', 'C', 'D']

    for i in range(5):
        for j in range(4):
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 1)

    for i in ['A', 'B', 'C', 'D']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i

def test_delete_column_last_name():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[4] = 123
    spreadsheet.width[3] = 321

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column('E')

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [50, 50, 50, 321]
    assert spreadsheet.headers == ['A', 'B', 'C', 'D']

    for i in range(5):
        for j in range(4):
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 1)

    for i in ['A', 'B', 'C', 'D']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i

def test_delete_column_middle_digit():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[2] = 222
    spreadsheet.width[1] = 123
    spreadsheet.width[3] = 321

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column(3)

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [50, 123, 321, 50]
    assert spreadsheet.headers == ['A', 'B', 'D', 'E']

    for i in range(5):
        for j in [0, 1]:
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 1)
        for j in [3, 4]:
            assert spreadsheet[j, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j, i + 1].value == (i + 1) * (j + 1)

    for i in ['A', 'B', 'D', 'E']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i

def test_delete_column_middle_name():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    columns = ['A', 'B', 'C', 'D', 'E']
    for i in columns:
        spreadsheet.style[i + '1'] = i
        spreadsheet.style[i + '2'] = i

    spreadsheet.width[2] = 222
    spreadsheet.width[1] = 123
    spreadsheet.width[3] = 321

    assert len(spreadsheet.style) == 25
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert len(spreadsheet.width) == 5
    assert len(spreadsheet.headers) == 5

    spreadsheet.delete_column('C')

    assert len(spreadsheet.style) == 20
    assert len(spreadsheet.data) == 5
    assert len(spreadsheet.values) == 5
    assert spreadsheet.width == [50, 123, 321, 50]
    assert spreadsheet.headers == ['A', 'B', 'D', 'E']

    for i in range(5):
        for j in [0, 1]:
            assert spreadsheet[j + 1, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j + 1, i + 1].value == (i + 1) * (j + 1)
        for j in [3, 4]:
            assert spreadsheet[j, i + 1].formula == (i + 1) * (j + 1)
            assert spreadsheet[j, i + 1].value == (i + 1) * (j + 1)

    for i in ['A', 'B', 'D', 'E']:
        assert spreadsheet[i, 1].style == i
        assert spreadsheet[i, 2].style == i

def test_get_formulas():
    spreadsheet = Spreadsheet(5, 5)

    for i in range(5):
        for j in range(5):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    # get_formula returns deep copy
    formulas = spreadsheet.get_formulas()
    formulas[0][0] = 0

    assert spreadsheet[1, 1].formula == 1
    assert formulas[0][0] == 0

def test_get_df():
    spreadsheet = Spreadsheet(3, 3)

    for i in range(3):
        for j in range(3):
            spreadsheet.data[i][j] = (i + 1) * (j + 1)
            spreadsheet.values[i][j] = (i + 1) * (j + 1)

    # get_formula returns deep copy
    df = spreadsheet.df('formulas')

    expected_df = DataFrame([[1,2,3],[2,4,6],[3,6,9]], columns=['A', 'B', 'C'], index=[1,2,3])
    assert_frame_equal(expected_df, df)

def test_deserialize_latin1():
    reformatter = PropertyReformatter(None)

    spreadsheet = reformatter.to_spreadsheet(SPREADSHEET_LATIN1)

    assert spreadsheet.cell(2, 1).formula == 'using ° or umlauts (ä,ö,ü)'



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import datetime
import time
import uuid

import pytest


def test_create_delete_experiment(space):
    o = space.openbis
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    new_code = 'test_experiment_' + timestamp

    with pytest.raises(TypeError):
        # experiments must be assigned to a project
        e_new = o.new_experiment(
            code=new_code,
            type='UNKNOWN',
        )

    project = o.get_projects()[0]

    e_new = o.new_experiment(
        code=new_code,
        project=project,
        type='UNKNOWN',
    )
    assert e_new.project is not None
    assert e_new.permId == ''

    e_new.save()

    assert e_new.permId is not None
    assert e_new.code == new_code.upper()

    e_exists = o.get_experiment(e_new.permId)
    assert e_exists is not None

    e_new.delete('delete test experiment ' + new_code.upper())

    with pytest.raises(ValueError):
        e_no_longer_exists = o.get_experiment(e_exists.permId)


def test_revert_deletion_experiment(space):
    o = space.openbis
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper() + "_" + str(uuid.uuid4())
    new_code = 'test_experiment_' + timestamp

    project = o.get_projects()[0]

    e_new = o.new_experiment(
        code=new_code,
        project=project,
        type='UNKNOWN',
    )
    assert e_new.project is not None
    assert e_new.permId == ''

    e_new.save()

    assert e_new.permId is not None
    assert e_new.code == new_code.upper()

    e_exists = o.get_experiment(e_new.permId)
    assert e_exists is not None

    e_new.delete('delete test experiment ' + new_code.upper())

    with pytest.raises(ValueError):
        e_no_longer_exists = o.get_experiment(e_exists.permId)
        assert "This should fail" is None

    df = o.get_deletions()
    assert df[df['permId'] == e_exists.permId].empty is False
    deletionId = df[df['permId'] == e_exists.permId]['deletionId'].iloc[0]
    o.revert_deletions([deletionId])

    e_exists = o.get_experiment(e_exists.permId)
    assert e_exists is not None
    df = o.get_deletions()
    if df.empty is False:
        assert df[df['permId'] == e_exists.permId].empty is True


def test_get_experiments(space):
    # test paging
    o = space.openbis
    current_datasets = o.get_experiments(start_with=1, count=1)
    assert current_datasets is not None
    # we cannot assert == 1, because search is delayed due to lucene search...
    assert len(current_datasets) <= 1


def test_experiment_property_in_isoformat_date(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").lower()

    # Create custom TIMESTAMP property type
    property_type_code = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type timestamp for experiment',
        description='custom property created in unit test',
        dataType='TIMESTAMP',
    )
    pt_date.save()

    type_code = "test_experiment_type_" + timestamp + "_" + str(uuid.uuid4())
    experiment_type = o.new_experiment_type(
        type_code,
        description=None,
        validationPlugin=None,
    )
    experiment_type.save()
    experiment_type.assign_property(property_type_code)

    project = o.get_projects()[0]
    code = "my_experiment_{}".format(timestamp)
    timestamp_property = datetime.datetime.now().isoformat()
    props = {property_type_code: timestamp_property}

    exp = o.new_experiment(code=code, project=project, type=type_code, props=props)
    exp.save()

    # New experiment case
    assert len(exp.p()) == 1
    assert exp.p[property_type_code] is not None

    # Update experiment case
    exp.p[property_type_code] = timestamp_property
    exp.save()

    assert len(exp.p()) == 1
    assert exp.p[property_type_code] is not None




#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import random

import time
import uuid


def test_create_delete_vocabulary(openbis_instance):
    o = openbis_instance
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    voc_code = 'test_voc_' + timestamp + "_" + str(random.randint(0, 1000))

    voc = o.new_vocabulary(
        code=voc_code,
        description='description of vocabulary',
        urlTemplate='https://ethz.ch',
        terms=[
            {"code": 'term_code1', "label": "term_label1", "description": "term_description1"},
            {"code": 'term_code2', "label": "term_label2", "description": "term_description2"},
            {"code": 'term_code3', "label": "term_label3", "description": "term_description3"}
        ],
        chosenFromList=False
    )
    assert voc.registrationDate is None
    voc.save()
    assert voc is not None
    assert voc.registrationDate is not None
    assert voc.chosenFromList is False

    voc_exists = o.get_vocabulary(voc_code)
    assert voc_exists is not None
    assert voc_exists.code == voc_code.upper()

    voc.description = 'description changed'
    voc.chosenFromList = True
    voc.save()

    assert voc.description == 'description changed'
    assert voc.chosenFromList is True

    voc.delete('test on ' + str(timestamp))


def test_create_vocabulary_and_delete_term(openbis_instance):
    o = openbis_instance
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    voc_code = 'test_voc_' + timestamp + "_" + str(random.randint(0, 1000))

    voc = o.new_vocabulary(
        code=voc_code,
        description='description of vocabulary',
        urlTemplate='https://ethz.ch',
        terms=[
            {"code": 'term_code1', "label": "term_label1", "description": "term_description1"},
            {"code": 'term_code2', "label": "term_label2", "description": "term_description2"},
            {"code": 'term_code3', "label": "term_label3", "description": "term_description3"}
        ],
        chosenFromList=False
    )
    assert voc.registrationDate is None
    voc.save()
    assert voc is not None
    assert voc.registrationDate is not None
    assert voc.chosenFromList is False

    term = o.get_term('term_code3'.upper(), voc_code.upper())
    term.delete('some_reason')

    saved_vocab = o.get_vocabulary(voc_code.upper())
    terms = saved_vocab.get_terms()
    assert list(terms.df['code']) == ['TERM_CODE1', 'TERM_CODE2']


def test_create_set_and_remove_vocabulary_property(space):
    o = space.openbis
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    voc_code = 'test_voc_' + timestamp + "_" + str(random.randint(0, 1000))

    voc = o.new_vocabulary(
        code=voc_code,
        description='description of vocabulary',
        urlTemplate='https://ethz.ch',
        terms=[
            {"code": 'term_code1', "label": "term_label1", "description": "term_description1"},
            {"code": 'term_code2', "label": "term_label2", "description": "term_description2"},
            {"code": 'term_code3', "label": "term_label3", "description": "term_description3"}
        ],
        chosenFromList=False
    )
    voc.save()

    voc_exists = o.get_vocabulary(voc_code)
    assert voc_exists is not None
    assert voc_exists.code == voc_code.upper()

    property_type_code = "test_vocabulary_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type controlled vocabulary',
        description='custom property created in unit test',
        dataType='CONTROLLEDVOCABULARY',
        vocabulary=voc_code
    )
    pt_date.save()

    # Create custom sample type
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Assign created property to new sample type
    sample_type.assign_property(
        prop=property_type_code,
        section='vocabulary section',
        ordinal=1,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    project = space.new_project(f'my_project_{timestamp}')
    project.save()

    collection = o.new_experiment('UNKNOWN', f'my_collection_{timestamp}', project)
    collection.save()

    sample_code = "my_sample_{}".format(timestamp)

    sample = space.new_sample(code=sample_code,
                              type=sample_type_code,
                              experiment=collection,
                              props={
                                  property_type_code.lower(): 'term_code1'})
    sample.save()

    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key.lower() == property_type_code.lower()
    assert val == 'term_code1'.upper()

    sample.props[property_type_code.lower()] = None
    sample.save()

    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key.lower() == property_type_code.lower()
    assert val is None


def test_create_delete_vocabulary_term(openbis_instance):
    o = openbis_instance
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    voc_code = 'test_voc_' + timestamp + "_" + str(random.randint(0, 1000))

    voc = o.new_vocabulary(
        code=voc_code,
        description='description of vocabulary',
        urlTemplate='https://ethz.ch',
        terms=[
            {"code": 'term_code1', "label": "term_label1", "description": "term_description1"},
            {"code": 'term_code2', "label": "term_label2", "description": "term_description2"},
            {"code": 'term_code3', "label": "term_label3", "description": "term_description3"}
        ],
        chosenFromList=False
    )
    assert voc.registrationDate is None
    voc.save()
    assert voc is not None
    assert voc.registrationDate is not None
    assert voc.chosenFromList is False

    voc_exists = o.get_vocabulary(voc_code)
    assert voc_exists is not None
    assert voc_exists.code == voc_code.upper()

    # Create vocabulary term
    voc_term = o.new_term('term_code4', voc_code.upper(), 'term_label4', 'term_description4')
    voc_term.save()

    assert voc_term is not None
    assert voc_term.registrationDate is not None
    voc_terms = o.get_terms(voc_code, use_cache=False)
    assert len(voc_terms) == 4

    # Delete vocabulary term
    voc_term.delete('some test reason')

    voc_terms = o.get_terms(voc_code, use_cache=False)
    assert len(voc_terms) == 3




#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import datetime
import random
import re
import time
import uuid

import pandas as pd
import pytest


def test_create_delete_sample(space):
    o = space.openbis

    sample_type = "UNKNOWN"
    sample = o.new_sample(
        code="illegal sample name with spaces", type=sample_type, space=space
    )
    with pytest.raises(ValueError):
        sample.save()
        assert "should not have been created" is None

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_code = "test_sample_" + timestamp + "_" + str(random.randint(0, 1000))
    sample = o.new_sample(code=sample_code, type=sample_type, space=space)
    assert sample is not None
    assert sample.space.code == space.code
    assert sample.code == sample_code
    assert sample.permId == ""
    sample.save()

    # now there should appear a permId
    assert sample.permId is not None

    # get it by permId
    sample_by_permId = o.get_sample(sample.permId)
    assert sample_by_permId is not None

    sample_by_permId = space.get_sample(sample.permId)
    assert sample_by_permId is not None

    assert sample_by_permId.registrator is not None
    assert sample_by_permId.registrationDate is not None
    # check date format: 2019-03-22 11:36:40
    assert (
            re.search(
                r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2}$",
                sample_by_permId.registrationDate,
            )
            is not None
    )

    # get sample by identifier
    sample_by_identifier = o.get_sample(sample.identifier)
    assert sample_by_identifier is not None

    sample_by_identifier = space.get_sample(sample.identifier)
    assert sample_by_identifier is not None

    sample.delete("sample creation test on " + timestamp)

    with pytest.raises(ValueError):
        o.get_sample(sample.permId)
        assert "This should fail" is None


def test_revert_deletion_sample(space):
    o = space.openbis

    sample_type = "UNKNOWN"
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_code = "test_sample_" + timestamp + "_" + str(random.randint(0, 1000))
    sample = o.new_sample(code=sample_code, type=sample_type, space=space)
    assert sample is not None
    assert sample.space.code == space.code
    assert sample.code == sample_code
    assert sample.permId == ""
    sample.save()

    # now there should appear a permId
    assert sample.permId is not None

    # get it by permId
    sample_by_permId = o.get_sample(sample.permId)
    assert sample_by_permId is not None

    sample_by_permId = space.get_sample(sample.permId)
    assert sample_by_permId is not None

    assert sample_by_permId.registrator is not None
    assert sample_by_permId.registrationDate is not None
    # check date format: 2019-03-22 11:36:40
    assert (
            re.search(
                r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2}$",
                sample_by_permId.registrationDate,
            )
            is not None
    )

    sample.delete("sample creation test on " + timestamp)

    with pytest.raises(ValueError):
        o.get_sample(sample.permId)
        assert "This should fail" is None

    df = o.get_deletions()
    assert df[df['permId'] == sample.permId].empty is False
    deletionId = df[df['permId'] == sample.permId]['deletionId'].iloc[0]
    o.revert_deletions([deletionId])

    sample = o.get_sample(sample.permId)
    assert sample is not None
    df = o.get_deletions()
    if df.empty is False:
        assert df[df['permId'] == sample.permId].empty is True


def test_create_delete_space_sample(space):
    o = space.openbis
    sample_type = "UNKNOWN"
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_code = "test_sample_" + timestamp + "_" + str(random.randint(0, 1000))

    sample = space.new_sample(code=sample_code, type=sample_type)
    assert sample is not None
    assert sample.space.code == space.code
    assert sample.code == sample_code
    sample.save()
    assert sample.permId is not None
    sample.delete("sample space creation test on " + timestamp)


def test_parent_child(space):
    o = space.openbis
    sample_type = "UNKNOWN"
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    parent_code = (
            "parent_sample_{}".format(timestamp) + "_" + str(random.randint(0, 1000))
    )
    sample_parent = o.new_sample(code=parent_code, type=sample_type, space=space)
    sample_parent.save()

    child_code = "child_sample_{}".format(timestamp)
    sample_child = o.new_sample(
        code=child_code, type=sample_type, space=space, parent=sample_parent
    )
    sample_child.save()
    time.sleep(5)

    ex_sample_parents = sample_child.get_parents()
    ex_sample_parent = ex_sample_parents[0]
    assert (
            ex_sample_parent.identifier == "/{}/{}".format(space.code, parent_code).upper()
    )

    ex_sample_children = ex_sample_parent.get_children()
    ex_sample_child = ex_sample_children[0]
    assert ex_sample_child.identifier == "/{}/{}".format(space.code, child_code).upper()

    sample_parent.delete("sample parent-child creation test on " + timestamp)
    sample_child.delete("sample parent-child creation test on " + timestamp)


def test_empty_data_frame(openbis_instance):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())

    sample_type = openbis_instance.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        subcodeUnique=False,
        listable=True,
        showContainer=False,
        showParents=True,
        showParentMetadata=False
    )
    sample_type.save()

    s = openbis_instance.get_sample_type(sample_type_code)
    pa = s.get_property_assignments()

    attrs = [
        "code",
        "dataType",
        "section",
        "ordinal",
        "mandatory",
        "initialValueForExistingEntities",
        "showInEditView",
        "showRawValueInForms",
        "registrator",
        "registrationDate",
        "plugin"
    ]

    pd.testing.assert_frame_equal(pa.df, pd.DataFrame(columns=attrs))


def test_new_sample_type_with_description(openbis_instance):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())

    description = 'my test description'

    sample_type = openbis_instance.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        subcodeUnique=False,
        listable=True,
        showContainer=False,
        showParents=True,
        showParentMetadata=False,
        description=description
    )
    sample_type.save()

    s = openbis_instance.get_sample_type(sample_type_code)
    assert s.description == description


def test_sample_date_property(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    # Create custom DATE property type
    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='DATE',
    )
    pt_date.save()

    # Create custom sample type
    sample_type_code = "test_sample_type_" + current_date + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Assign created property to new sample type
    sample_type.assign_property(
        prop=property_type_code,
        section='',
        ordinal=1,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    sample_code = "my_sample_{}_{}".format(current_date, str(uuid.uuid4()))
    sample = o.new_sample(code=sample_code,
                          type=sample_type_code,
                          space=space,
                          props={
                              property_type_code: current_date})
    sample.save()

    # New item case
    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key == property_type_code
    assert val == current_date

    # Update item case
    sample.props = {property_type_code: '2024-05-16'}
    sample.save()

    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key == property_type_code
    assert val == '2024-05-16'


def test_sample_property_in_isoformat_timestamp(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").lower()

    # Create custom TIMESTAMP property type
    property_type_code = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type timestamp',
        description='custom property created in unit test',
        dataType='TIMESTAMP',
    )
    pt_date.save()

    # Create custom sample type
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Assign created property to new sample type
    sample_type.assign_property(
        prop=property_type_code,
        section='',
        ordinal=5,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    sample_code = "my_sample_{}".format(timestamp)
    # Create new sample with timestamp property in non-supported format
    timestamp_property = datetime.datetime.now().isoformat()
    sample = o.new_sample(code=sample_code,
                          type=sample_type_code,
                          space=space,
                          props={
                              property_type_code: timestamp_property})
    sample.save()

    # New item case
    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key == property_type_code

    # Update item case
    sample.props = {property_type_code: timestamp_property}
    sample.save()

    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key == property_type_code

def test_create_sample_type_assign_property(space):
    name_suffix = str(time.time())
    sc = "TEST_" + name_suffix
    pc = "ESFA_" + name_suffix
    ptc1 = "START_DATE_" + name_suffix
    ptc2 = "EXP_DESCRIPTION_" + name_suffix
    stc = "EXPERIMENTAL_STEP_MILAR_" + name_suffix

    # Create the new space and project
    sp = space.openbis.new_space(code=sc, description="Test space")
    sp.save()
    pr = space.openbis.new_project(code=pc, space=sc, description="ESFA experiments")
    pr.save()

    # Create the experiment
    exp = space.openbis.new_collection(code=pc, project="/" + sc + "/" + pc, type="COLLECTION")
    exp.save()

    # Create the sample type
    date_prop = space.openbis.new_property_type(code=ptc1, dataType="TIMESTAMP",
                                                label="Start date",
                                                description="Date of the measurement")
    date_prop.save()
    date_prop = space.openbis.new_property_type(code=ptc2, dataType="MULTILINE_VARCHAR",
                                                label="Experimental description",
                                                description="Experimental description")
    date_prop.save()
    st = space.openbis.new_sample_type(code=stc, generatedCodePrefix="EXSTEPMILAR")
    st.save()

    if st is None:
        print(space.openbis.get_sample_types())
        st = space.openbis.get_sample_type(stc)
        st.save()

    st.assign_property(ptc1)
    st.assign_property(ptc2)
    st.assign_property("$NAME")
    st.save()


def test_del_child_from_sample(space):
    # Prepare
    sample_type = "UNKNOWN"
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    sample_code1 = "test_sample_child_" + timestamp + "_" + str(random.randint(0, 1000))

    sample_child = space.new_sample(code=sample_code1, type=sample_type)
    sample_child.save()

    sample_code2 = "test_sample_parent_" + timestamp + "_" + str(random.randint(0, 1000))
    sample_parent = space.new_sample(code=sample_code2, type=sample_type)
    sample_parent.children = [sample_child]
    sample_parent.save()

    assert sample_parent.children == [sample_child.identifier]

    # Act & Assert
    item = str(sample_parent.children[0])
    sample_parent.del_children(item)
    sample_parent.save()

    assert sample_parent.children == []


def test_sample_with_auto_generated_code(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").lower()

    # Create custom VARCHAR property type
    property_type_code = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type varchar',
        description='custom property created in unit test',
        dataType='VARCHAR',
    )
    pt_date.save()

    # Create custom sample type
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Assign created property to new sample type
    sample_type.assign_property(
        prop=property_type_code,
        section='',
        ordinal=5,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    project = space.new_project(f'my_project_{timestamp}')
    project.save()

    collection = o.new_experiment('UNKNOWN', f'my_collection_{timestamp}', project)
    collection.save()

    sample_code = "my_sample_{}".format(timestamp)
    # create varchar property
    timestamp_property = str(datetime.datetime.now().isoformat())

    sample = space.new_sample(code=sample_code,
                          type=sample_type_code,
                          experiment=collection,
                          props={
                              property_type_code: timestamp_property})
    sample.save()

    assert len(sample.props()) == 1
    key, val = sample.props().popitem()
    assert key == property_type_code
    assert val == timestamp_property


def test_create_sample_clear_property_values(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").lower()

    # Create custom Integer property type
    property_type_code_1 = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_integer = o.new_property_type(
        code=property_type_code_1,
        label='custom property of INTEGER type',
        description='custom property created in unit test',
        dataType='INTEGER'
    )
    pt_integer.save()

    property_type_code_2 = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_integer_2 = o.new_property_type(
        code=property_type_code_2,
        label='custom property of BOOLEAN type',
        description='custom property created in unit test',
        dataType='BOOLEAN'
    )
    pt_integer_2.save()

    # Create custom sample type
    sample_type_code = "test_sample_type_" + timestamp + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Assign created property to new sample type
    sample_type.assign_property(
        prop=property_type_code_1,
        section='',
        ordinal=1,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    sample_type.assign_property(
        prop=property_type_code_2,
        section='',
        ordinal=2,
        mandatory=False,
        showInEditView=True,
        showRawValueInForms=True
    )

    sample_code = "my_sample_{}".format(timestamp)

    # Create new sample with timestamp property in non-supported format
    sample = o.new_sample(code=sample_code,
                          type=sample_type_code,
                          space=space,
                          props={
                              property_type_code_1: 1,
                              property_type_code_2: True,
                                })
    sample.save()

    # New item case
    assert len(sample.props()) == 2
    key, val = sample.props().popitem()
    assert key in [property_type_code_1, property_type_code_2]
    assert val in [1, 'true']
    key, val = sample.props().popitem()
    assert key in [property_type_code_1, property_type_code_2]
    assert val in [1, 'true']

    # Update item case
    sample.props = {property_type_code_1: '', property_type_code_2: None}
    sample.save()

    assert len(sample.props()) == 2
    key, val = sample.props().popitem()
    assert key in [property_type_code_1, property_type_code_2]
    assert val is None
    key, val = sample.props().popitem()
    assert key in [property_type_code_1, property_type_code_2]
    assert val is None

def test_create_sample_with_spreadsheet(space):
    name_suffix = str(time.time())
    sc = "TEST_1_" + name_suffix
    pc = "ESFA_1_" + name_suffix
    ptc1 = "XML_SPREADSHEET_" + name_suffix
    stc = "EXPERIMENTAL_STEP_MILAR_" + name_suffix

    # Create the new space and project
    sp = space.openbis.new_space(code=sc, description="Test space")
    sp.save()
    pr = space.openbis.new_project(code=pc, space=sc, description="ESFA experiments")
    pr.save()

    # Create the experiment
    exp = space.openbis.new_collection(code=pc, project="/" + sc + "/" + pc, type="COLLECTION")
    exp.save()

    # Create the sample type
    date_prop = space.openbis.new_property_type(code=ptc1, dataType="XML",
                                                label="test spreadsheet",
                                                description="test xml property",
                                                metaData={"custom_widget": "Spreadsheet"})
    date_prop.save()

    st = space.openbis.new_sample_type(code=stc, generatedCodePrefix="TEST_PATTERN_")
    st.save()

    if st is None:
        print(space.openbis.get_sample_types())
        st = space.openbis.get_sample_type(stc)
        st.save()

    st.assign_property(ptc1)
    st.save()

    spreadsheet = space.openbis.new_spreadsheet(10, 10)
    spreadsheet.data[0][0] = 10

    sample = space.openbis.new_sample(code=f"CODE_{name_suffix}",
                                      type=st.code,
                                      experiment=exp,
                                      props={ptc1.lower(): spreadsheet})
    sample.save()

    assert sample.props[ptc1.lower()] is not None
    assert sample.props[ptc1.lower()].data[0][0] == 10

def test_parent_child_in_project(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()

    project = space.new_project(f'my_project_{timestamp}')
    project.save()

    collection = o.new_experiment('UNKNOWN', f'my_collection_{timestamp}', project)
    collection.save()

    sample_type = "UNKNOWN"

    parent_code = (
            "parent_sample_{}".format(timestamp) + "_" + str(random.randint(0, 1000))
    )
    sample_parent = o.new_sample(code=parent_code, type=sample_type, collection=collection)
    sample_parent.save()

    child_code = "child_sample_{}".format(timestamp)
    sample_child = o.new_sample(
        code=child_code, type=sample_type, collection=collection, parent=sample_parent
    )
    sample_child.save()
    time.sleep(5)

    samples = o.get_samples(sample_parent)
    assert len(samples) == 1
    assert samples[0].children == '--NOT FETCHED--'
    child = samples[0].get_children()
    assert child is not None
    assert len(child) == 1
    assert samples[0].children[0] == sample_child.identifier
    children = samples.get_children()
    assert children is not None
    assert len(children) == 1
    assert samples[0].get_children().get_parents()[0].identifier == samples[0].identifier



#   Copyright ETH 2019 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import uuid

import pytest


def test_create_delete_plugin(openbis_instance):
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    pl_name = 'test_plugin_' + timestamp

    plugin = openbis_instance.new_plugin(
	name= pl_name,
	pluginType='ENTITY_VALIDATION',   # or 'DYNAMIC_PROPERTY' or 'MANAGED_PROPERTY',
	entityKind = None,                # or 'SAMPLE', 'MATERIAL', 'EXPERIMENT', 'DATA_SET'
	script = 'def calculate(): pass'  # JYTHON script
    )

    with pytest.raises(ValueError):
        plugin.pluginType = 'Rubbish'
        assert 1 == 0

    with pytest.raises(ValueError):
        plugin.entityKind = 'even more Rubbish'
        assert 1 == 0

    plugin.save()
    pl_exists = openbis_instance.get_plugin(pl_name)
    assert pl_exists is not None

    plugin.delete("test on {}".format(timestamp))

    with pytest.raises(ValueError):
        pl_not_exists = openbis_instance.get_plugin(plugin.name)
        assert pl_not_exists is None


def test_search_plugins(openbis_instance):
    plugins = openbis_instance.get_plugins(start_with=1, count=1)
    assert len(plugins) == 1
    assert plugins.df.__class__.__name__ == 'DataFrame'
    assert len(plugins) == len(plugins.df)

    plugin = plugins[0]
    assert plugin.__class__.__name__ == 'Plugin'


def test_assign_plugin_to_sample_type(openbis_instance):

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()

    plugin_name = 'TEST.NEW_PLUGIN_' + timestamp + "_" + str(uuid.uuid4())
    pl = openbis_instance.new_plugin(
        name=plugin_name,
        pluginType='DYNAMIC_PROPERTY',
        entityKind='SAMPLE',
        script='def validate(x, True):\n\treturn "OK"'
    )
    pl.save()

    sample_type_code = 'SOME_TEST_SAMPLE_TYPE_' + timestamp + "_" + str(uuid.uuid4())
    sample_type = openbis_instance.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix='STST-',
        autoGeneratedCode=True
    )
    sample_type.save()

    sample_type = openbis_instance.get_sample_type(sample_type_code)
    assignments = sample_type.get_property_assignments().df

    assert len(assignments) == 0

    property_code = 'SOME_TEST_PROPERTY_' + timestamp + "_" + str(uuid.uuid4())
    prop_type = openbis_instance.new_property_type(
        code=property_code,
        label='test property',
        description='test property',
        dataType='VARCHAR',
    )
    prop_type.save()

    sample_type.assign_property(prop_type, section='General info', ordinal=2,
                                initialValueForExistingEntities='', plugin=plugin_name)

    assignments = sample_type.get_property_assignments().df

    assert len(assignments) == 1
    assert assignments.loc[0]['plugin'] == plugin_name

    sample = openbis_instance.new_sample(
        type=sample_type.code,
        space='DEFAULT',
        experiment='/DEFAULT/DEFAULT/DEFAULT',
        props={property_code.lower(): "some_value"}
    )
    sample.save()

    sample_type.revoke_property(property_code, force=True)

    assignments = sample_type.get_property_assignments().df
    assert len(assignments) == 0


def test_update_plugin(openbis_instance):
    timestamp = time.strftime('%a_%y%m%d_%H%M%S').upper()
    pl_name = 'test_plugin_' + timestamp

    plugin = openbis_instance.new_plugin(
        name=pl_name,
        pluginType='ENTITY_VALIDATION',  # or 'DYNAMIC_PROPERTY' or 'MANAGED_PROPERTY',
        entityKind=None,  # or 'SAMPLE', 'MATERIAL', 'EXPERIMENT', 'DATA_SET'
        script='def calculate(): pass'  # JYTHON script
    )
    plugin.save()
    pl_exists = openbis_instance.get_plugin(pl_name)
    assert pl_exists is not None

    plugin.description = 'some boring description'
    plugin.available = True
    plugin.script = 'def calculate(): fail'
    plugin.save()

    pl_exists = openbis_instance.get_plugin(pl_name)
    assert pl_exists is not None
    assert pl_exists.description == 'some boring description'
    assert pl_exists.script == 'def calculate(): fail'
    assert pl_exists.available is True




#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import random
import re

import pytest
import time
from random import randint


def test_crud_tag(openbis_instance):
    tag_name = "test_tag_{}".format(randint(0, 1000)).upper()
    description = "description of tag " + tag_name

    tag = openbis_instance.new_tag(code=tag_name, description=description)

    assert tag.code == tag_name
    assert tag.description == description
    assert tag.permId == ""

    tag.save()

    assert tag.permId is not None

    tag_exists = openbis_instance.get_tag(tag.permId)
    assert tag_exists is not None

    tag_by_code = openbis_instance.get_tag(tag.code)
    assert tag_by_code is not None
    assert tag_by_code.permId == tag_exists.permId

    altered_description = "altered description of tag " + tag_name
    tag.description = altered_description
    tag.save()
    assert tag.description == altered_description

    tag.delete("test")

    with pytest.raises(ValueError):
        tag_does_not_exists = openbis_instance.get_tag(tag.permId, use_cache=False)


def test_get_tags(openbis_instance):
    tags = openbis_instance.get_tags()
    assert tags is not None
    assert tags.__class__.__name__ == "Things"
    assert tags.df.__class__.__name__ == "DataFrame"

    if len(tags) > 0:
        assert tags[0].__class__.__name__ == "Tag"

    if len(tags) > 1:
        tag1 = tags[0]
        tag2 = tags[1]

        tag_coll = openbis_instance.get_tag([tag1.permId, tag2.permId])
        assert len(tag_coll) == 2
        assert tag_coll.__class__.__name__ == "Things"
        assert tag_coll.df.__class__.__name__ == "DataFrame"



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import random
import re

import pytest
import time
from pybis import DataSet
from pybis import Openbis


def test_token(openbis_instance):
    assert openbis_instance.token is not None
    assert openbis_instance.is_token_valid(openbis_instance.token) is True
    assert openbis_instance.is_session_active() is True


### Temporarily disabled
# def test_http_only(openbis_instance):
#     with pytest.raises(Exception):
#         new_instance = Openbis("http://localhost")
#         assert new_instance is None
#
#     new_instance = Openbis(
#         url="http://localhost",
#         allow_http_but_do_not_use_this_in_production_and_only_within_safe_networks=True,
#     )
#     assert new_instance is not None


def test_cached_token(other_openbis_instance):
    assert other_openbis_instance.is_token_valid() is True

    other_openbis_instance.logout()
    assert other_openbis_instance.is_token_valid() is False


def test_create_permId(openbis_instance):
    permId = openbis_instance.create_permId()
    assert permId is not None
    m = re.search("([0-9]){17}-([0-9]*)", permId)
    ts = m.group(0)
    assert ts is not None
    count = m.group(1)
    assert count is not None

def test_openbis():
    def get_instance():
        base_url = "http://localhost:8888/openbis-test"
        # base_url = "https://localhost:8443/openbis"
        # base_url = "https://openbis-sis-ci-sprint.ethz.ch/"
        # base_url = "https://local.openbis.ch/openbis"
        openbis_instance = Openbis(
            url=base_url,
            verify_certificates=False,
            allow_http_but_do_not_use_this_in_production_and_only_within_safe_networks=True
        )
        token = openbis_instance.login('admin', 'changeit')
        print(token)
        return openbis_instance

    o = get_instance()



    print("DONE")


#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .property import PropertyHolder
from .property_reformatter import PropertyReformatter

from urllib.parse import quote
import copy

class Sample(OpenBisObject, entity="sample", single_item_method_name="get_sample"):
    """A Sample (new: Object) is one of the most commonly used entities in openBIS."""

    def __init__(
            self, openbis_obj, type, project=None, data=None, props=None, attrs=None, **kwargs
    ):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["type"] = type
        ph = PropertyHolder(openbis_obj, type)
        self.__dict__["p"] = ph
        self.__dict__["props"] = ph
        self.__dict__["a"] = AttrHolder(openbis_obj, "sample", type)
        self.__dict__["formatter"] = PropertyReformatter(openbis_obj)

        if data is not None:
            self._set_data(data)

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

            if "experiment" in kwargs:
                try:
                    experiment = self.experiment
                    if not "space" in kwargs:
                        project = experiment.project
                        self.a.space = project.space
                except Exception:
                    pass

        if project is None:
            if self.experiment:
                self.project = self.experiment.project
        else:
            self.project = project

        if props is not None:
            for key in props:
                # self.p[key] = props[key]
                setattr(self.p, key, props[key])

        if getattr(self, "parents") is None:
            self.a.__dict__["_parents"] = []
        else:
            if not self.is_new:
                if attrs is not None and "parents" not in attrs:
                    self.a.__dict__["_parents"] = None
                else:
                    self.a.__dict__["_parents_orig"] = copy.copy(self.a.__dict__["_parents"])

        if getattr(self, "children") is None:
            self.a.__dict__["_children"] = []
        else:
            if not self.is_new:
                if attrs is not None and "children" not in attrs:
                    self.a.__dict__["_children"] = None
                else:
                    self.a.__dict__["_children_orig"] = copy.copy(self.a.__dict__["_children"])

        if getattr(self, "components") is None:
            self.a.__dict__["_components"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_components_orig"] = self.a.__dict__["_components"]

    def _set_data(self, data):
        # assign the attribute data to self.a by calling it
        # (invoking the AttrHolder.__call__ function)
        self.a(data)
        self.__dict__["data"] = data

        # put the properties in the self.p namespace
        for key, value in data["properties"].items():
            property_type = self.p._property_names[key.lower()]
            data_type = property_type['dataType']
            if "multiValue" in property_type:
                if property_type['multiValue'] is True:
                    if type(value) is not list:
                        value = [value]
                    if data_type in ("ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        value = [self.formatter.to_array(data_type, x) for x in value]
                    else:
                        value = self.formatter.to_array(data_type, value)
                else:
                    if type(value) is list and data_type not in ("ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        raise ValueError(f'Property type {property_type} is not a multi-value property!')
                    if data_type in ("ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                        value = self.formatter.to_array(data_type, value)
            else:
                if data_type in ("ARRAY_INTEGER", "ARRAY_REAL", "ARRAY_STRING", "ARRAY_TIMESTAMP"):
                    value = self.formatter.to_array(data_type, value)
            if (data_type == 'XML' and 'metaData' in property_type and 'custom_widget' in property_type['metaData']
                    and property_type['metaData']['custom_widget'].upper() == 'SPREADSHEET'):
                    if key.lower() in self.p.__dict__:
                        old_spreadsheet = self.p.__dict__[key.lower()]
                        old_spreadsheet._set_data(self.formatter.to_spreadsheet(value))
                        value = old_spreadsheet
                    else:
                        value = self.formatter.to_spreadsheet(value)
            self.p.__dict__[key.lower()] = value

    def __dir__(self):
        return [
            "type",
            "get_parents()",
            "get_children()",
            "get_components()",
            "add_parents()",
            "add_children()",
            "add_components()",
            "del_parents()",
            "del_children()",
            "del_components()",
            "set_parents()",
            "set_children()",
            "set_components()",
            "get_datasets()",
            "space",
            "project",
            "experiment",
            "container",
            "tags",
            "set_tags()",
            "add_tags()",
            "del_tags()",
            "add_attachment()",
            "get_attachments()",
            "download_attachments()",
            "save()",
            "delete()",
            "mark_to_be_deleted()",
            "unmark_to_be_deleted()",
            "is_marked_to_be_deleted()",
            "attrs",
            "props",
        ] + super().__dir__()

    def _container(self, value=None):
        if value is not None:
            if value == "":
                if self.is_new:
                    pass
                else:
                    self.a.__dict__["_container"] = {}
            else:
                obj = None
                if isinstance(value, str):
                    # fetch object in openBIS, make sure it actually exists
                    obj = getattr(self._openbis, "get_sample")(value)
                elif value is None:
                    self.a.__dict__["_container"] = {}
                else:
                    obj = value

                self.a.__dict__["_container"] = obj.data["identifier"]

                # mark attribute as modified, if it's an existing entity
                if self.is_new:
                    pass
                else:
                    self.a.__dict__["_container"]["isModified"] = True
        else:
            try:
                return self.openbis.get_sample(self.a._container["identifier"])
            except Exception:
                pass

    @property
    def type(self):
        return self.__dict__["type"]

    @type.setter
    def type(self, type_name):
        sample_type = self.openbis.get_sample_type(type_name)
        self.p.__dict__["_type"] = sample_type
        self.a.__dict__["_type"] = sample_type

    def __getattr__(self, name):
        if name in ["container"]:
            return getattr(self, "_" + name)()

        return getattr(self.__dict__["a"], name)

    def __setattr__(self, name, value):
        if name in ["set_properties", "set_tags", "add_tags"]:
            raise ValueError("These are methods which should not be overwritten")

        if name in ["container"]:
            return getattr(self, "_" + name)(value)

        if name in ["p", "props"]:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__["p"], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            # must be an attribute in the AttributeHolder class
            setattr(self.__dict__["a"], name, value)

    def _repr_html_(self):
        return self.a._repr_html_()

    def __repr__(self):
        return self.a.__repr__()

    def set_properties(self, properties):
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties

    def get_datasets(self, **kwargs):
        return self.openbis.get_datasets(sample=self.permId, **kwargs)

    def get_projects(self, **kwargs):
        return self.openbis.get_project(withSamples=[self.permId], **kwargs)

    @property
    def experiment(self):
        try:
            return self.openbis.get_experiment(self._experiment["identifier"])
        except Exception:
            pass

    def save(self):
        """invoked when code is provided in cases when the type already generates
        the code automatically. In this case, we need to invoke the old V1 method.
        """
        if self.is_new and self.code is not None and self.type.autoGeneratedCode:
            request = self._new_attrs()
            if self.props:
                for prop_name, prop in self.props._property_names.items():
                    if prop["mandatory"]:
                        if (
                                getattr(self.props, prop_name) is None
                                or getattr(self.props, prop_name) == ""
                        ):
                            raise ValueError(
                                f"Property '{prop_name}' is mandatory and must not be None"
                            )
            properties = PropertyReformatter(self.openbis).format(self.props())

            for attr in request['params'][1][0]:
                if request['params'][1][0][attr] is not None and 'isModified' in request['params'][1][0][attr]:
                    del request['params'][1][0][attr]['isModified']

            request['params'][1][0]['properties'] = properties

            resp = self.openbis._post_request(self.openbis.as_v3, request)

            permId = resp[0]['permId']
            new_entity_data = self.openbis.get_sample(permId, only_data=True)
            self._set_data(new_entity_data)
            return self


        else:
            super().save()

    def get_eln_url(self):
        return f'{self.openbis.url}/webapp/eln-lims/?menuUniqueId=null&viewName=showViewSamplePageFromPermId&viewData={self.permId}'




#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


def test_admin_person_roles(openbis_instance):
    admin = openbis_instance.get_person(userId="admin")
    assert admin is not None

    # test role assignments
    roles = admin.get_roles()
    assert len(roles) > 0

    admin.assign_role("OBSERVER")

    roles = admin.get_roles()
    observer_role_exists = False
    for role in roles:
        if role.role == "OBSERVER":
            observer_role_exists = True
    assert observer_role_exists

    admin.assign_role(role="OBSERVER", space="DEFAULT")
    space_roles_exist = admin.get_roles(space="DEFAULT")
    assert len(space_roles_exist) == 1

    admin.revoke_role(role="OBSERVER")
    roles_exist = admin.get_roles()
    assert len(roles_exist) > 1

    admin.revoke_role(role="OBSERVER", space="DEFAULT")
    roles_exist = admin.get_roles()
    assert len(roles_exist) == 1



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import datetime
import os
import re
import time
import uuid

import pytest
import filecmp
import tempfile

from pybis.things import Things


def test_get_datasets_count(space):
    o = space.openbis
    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")
    dataset = o.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()

    try:
        current_datasets = o.get_datasets(count=1)
        assert current_datasets is not None
        assert len(current_datasets) == 1
    finally:
        dataset.delete("test_get_datasets_count", True)


def test_get_datasets_paging(space):
    o = space.openbis
    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")
    dataset1 = o.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset1.save()

    dataset2 = o.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset2.save()

    try:
        current_datasets = o.get_datasets(start_with=1, count=1)
        assert current_datasets is not None
        assert len(current_datasets) == 1
    finally:
        dataset1.delete("test_get_datasets_paging", True)
        dataset2.delete("test_get_datasets_paging", True)


def test_create_datasets_no_file(space):
    o = space.openbis
    with pytest.raises(Exception) as exc:
        o.new_dataset(
            type="RAW_DATA",
            experiment="/DEFAULT/DEFAULT/DEFAULT",
            props={"$name": "some good name"},
        )
    assert str(exc.value) == "please provide at least one file"


def test_create_delete_dataset(space):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    o = space.openbis
    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    dataset = o.new_dataset(
        type="RAW_DATA",
        sample="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name", "notes": "my notes"},
    )

    assert dataset is not None
    assert not dataset.permId
    assert dataset.p is not None
    assert dataset.p["$name"] == "some good name"
    assert dataset.p.notes == "my notes"

    with pytest.raises(Exception):
        dataset.non_existing_attribute = "invalid attribute"
        assert "attribute does not exist, should fail" is None

    with pytest.raises(Exception):
        dataset.p.non_existing_property = "invalid propery"
        assert "property does not exist, should fail" is None

    dataset.save()

    # now there should appear a permId in our object
    assert dataset.permId is not None
    permId = dataset.permId

    # get it by permId
    dataset_by_permId = o.get_dataset(dataset.permId)
    assert dataset_by_permId is not None
    assert dataset_by_permId.permId == permId
    assert dataset_by_permId.type is not None
    assert dataset_by_permId.type.code == "RAW_DATA"
    assert dataset_by_permId.kind == "PHYSICAL"
    assert dataset_by_permId.sample is not None
    assert dataset_by_permId.sample.code == "DEFAULT"
    assert dataset_by_permId.experiment is not None
    assert dataset_by_permId.experiment.code == "DEFAULT"

    assert dataset_by_permId.p is not None
    assert dataset_by_permId.p["$name"] == "some good name"
    assert dataset_by_permId.p.notes == "my notes"

    assert dataset_by_permId.registrator is not None
    assert dataset_by_permId.registrationDate is not None
    # check date format: 2019-03-22 11:36:40
    assert (
            re.search(
                r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2}$",
                dataset_by_permId.registrationDate,
            )
            is not None
    )

    # delete datasets
    dataset.delete("dataset creation test on " + timestamp, True)

    # Give openbis some time to process it
    time.sleep(1)
    # check that permanent deletion is working
    deletions = o.get_deletions(0, 10)
    assert len(deletions) == 0

    # get by permId should now throw an error
    with pytest.raises(Exception):
        deleted_ds = o.get_dataset(permId)


def test_revert_deletion_dataset(space):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    o = space.openbis
    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    dataset = o.new_dataset(
        type="RAW_DATA",
        sample="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name", "notes": "my notes"},
    )

    assert dataset is not None
    assert not dataset.permId
    assert dataset.p is not None
    assert dataset.p["$name"] == "some good name"
    assert dataset.p.notes == "my notes"

    dataset.save()

    # now there should appear a permId in our object
    assert dataset.permId is not None
    permId = dataset.permId

    # get it by permId
    dataset_by_permId = o.get_dataset(dataset.permId)
    assert dataset_by_permId is not None
    assert dataset_by_permId.permId == permId

    # delete datasets
    dataset.delete("dataset creation test on " + timestamp)

    # Give openbis some time to process it
    time.sleep(2)

    # get by permId should now throw an error
    with pytest.raises(Exception):
        deleted_ds = o.get_dataset(permId)
        assert "This should fail" is None

    df = o.get_deletions()
    assert df[df['permId'] == permId].empty is False
    deletionId = df[df['permId'] == permId]['deletionId'].iloc[0]
    o.revert_deletions([deletionId])

    dataset = o.get_dataset(permId)
    assert dataset is not None
    df = o.get_deletions()
    if df.empty is False:
        assert df[df['permId'] == permId].empty is True


def test_create_dataset_with_code(space):
    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    o = space.openbis

    dataset = o.new_dataset(
        type="UNKNOWN",
        code=timestamp,
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        sample="/DEFAULT/DEFAULT/DEFAULT",
        kind="CONTAINER",
    )

    assert dataset is not None
    assert not dataset.permId
    assert dataset.code == timestamp
    dataset.save()

    # our permId is now identical to the code we provided
    assert dataset.permId is not None
    assert dataset.permId == timestamp

    dataset.delete("dataset creation test on {}".format(timestamp))


def test_things_initialization(space):
    data_frame_result = [1, 2, 3]
    objects_result = [4, 5, 6]

    def create_data_frame(attrs, props, response):
        return data_frame_result

    def create_objects(response):
        return objects_result

    things = Things(
        openbis_obj=None,
        entity='dataset',
        identifier_name='permId',
        start_with=0,
        count=10,
        totalCount=10,
        response=None,
        df_initializer=create_data_frame,
        objects_initializer=create_objects
    )

    assert not things.is_df_initialised()
    assert not things.is_objects_initialised()

    assert things.df == data_frame_result
    assert things.objects == objects_result

    assert things.is_df_initialised()
    assert things.is_objects_initialised()


def test_create_new_dataset_v1(space):
    """Create dataset and upload file using upload scheme from before 3.6 api version"""
    openbis_instance = space.openbis

    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    # It is a hack to force old way of upload for testing.
    openbis_instance.get_server_information()._info["api-version"] = "3.5"

    dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()
    print(dataset)
    assert dataset.permId is not None
    assert dataset.file_list == ["original/testfile"]


def test_create_new_dataset_v3_single_file(space):
    openbis_instance = space.openbis

    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()
    print(dataset)
    assert dataset.permId is not None
    assert dataset.file_list == ["original/testfile"]


def test_create_new_dataset_v3_directory(space):
    openbis_instance = space.openbis

    testfile_path = os.path.join(os.path.dirname(__file__), "testdir")

    dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()
    print(dataset)
    assert dataset.permId is not None
    assert len(dataset.file_list) == 2
    assert dataset.file_list[0].endswith('testdir/testfile')


def test_dataset_property_in_isoformat_date(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").lower()

    # Create custom TIMESTAMP property type
    property_type_code = "test_property_type_" + timestamp + "_" + str(uuid.uuid4())
    pt_date = o.new_property_type(
        code=property_type_code,
        label='custom property of data type timestamp',
        description='custom property created in unit test',
        dataType='TIMESTAMP',
    )
    pt_date.save()

    # Create new dataset type
    type_code = "test_dataset_type_" + timestamp + "_" + str(uuid.uuid4())
    dataset_type = o.new_dataset_type(code=type_code)
    dataset_type.save()

    # Assign created property to new dataset type
    dataset_type.assign_property(property_type_code)

    # Create new dataset with timestamp property in non-supported format
    timestamp_property = datetime.datetime.now().isoformat()
    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    dataset = o.new_dataset(
        type=type_code,
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={property_type_code: timestamp_property},
    )
    dataset.save()

    # New dataset case
    assert len(dataset.p()) == 1
    assert dataset.p[property_type_code] is not None

    # Update dataset case
    dataset.p[property_type_code] = timestamp_property
    dataset.save()

    assert len(dataset.p()) == 1
    assert dataset.p[property_type_code] is not None

def test_create_new_dataset_with_parent(space):
    openbis_instance = space.openbis

    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()

    assert dataset.permId is not None
    assert dataset.file_list == ["original/testfile"]

    new_dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
        parents=[dataset.permId]
    )
    new_dataset.save()

    assert new_dataset.permId is not None
    assert new_dataset.file_list == ["original/testfile"]
    assert new_dataset.parents == [dataset.permId]

def test_create_new_dataset_in_chunks(space):
    openbis_instance = space.openbis

    testfile_path = os.path.join(os.path.dirname(__file__), "testdir/testfile_big")

    dataset = openbis_instance.new_dataset(
        type="RAW_DATA",
        experiment="/DEFAULT/DEFAULT/DEFAULT",
        files=[testfile_path],
        props={"$name": "some good name"},
    )
    dataset.save()

    assert dataset.permId is not None
    assert dataset.file_list == ["original/testfile_big"]

    with tempfile.TemporaryDirectory() as tmpdirname:
        dataset.download(destination=tmpdirname, create_default_folders=False)
        assert filecmp.cmp(testfile_path, os.path.join(tmpdirname, "testfile_big"))




#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
import random
import time

import pytest


def test_create_delete_project(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    project = o.new_project(space=space, code="illegal title contains spaces")
    with pytest.raises(ValueError):
        project.save()
        assert "should not have been created" is None

    project_name = "project_" + timestamp + "_" + str(random.randint(0, 1000))
    project = o.new_project(space=space, code=project_name)
    project.save()

    project_exists = o.get_project(project_name)
    assert project_exists is not None

    projects_exist = o.get_projects()
    assert len(projects_exist) > 0
    first_project = projects_exist[0]
    assert first_project is not None

    project_exists.delete("test project on " + timestamp)

    with pytest.raises(ValueError):
        project_no_longer_exists = o.get_project(project_name, use_cache=False)


def test_create_project_with_attachment(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()
    project_name = "project_" + timestamp + "_" + str(random.randint(0, 1000))
    filename = os.path.join(os.path.dirname(__file__), "testdir/testfile")

    if not os.path.exists(filename):
        raise ValueError("File not found: {}".format(filename))

    project = o.new_project(space=space, code=project_name, attachments=filename)
    assert project.attachments is not None
    project.save()

    project_exists = o.get_project(project_name)
    assert project_exists is not None
    assert project_exists.attachments is not None


def test_get_project_by_code(space):
    o = space.openbis

    timestamp = time.strftime("%a_%y%m%d_%H%M%S").upper()

    space_code_1 = "space_1_" + timestamp + "_" + str(random.randint(0, 1000))
    project_code = "project_" + timestamp

    o.new_space(code=space_code_1).save()

    o.new_project(space=space_code_1, code=project_code).save()
    project_exists = o.get_project(project_code)
    assert project_exists is not None




#   Copyright ETH 2025 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import time
import uuid

import pytest


def test_new_object_property_type(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    # Create custom SAMPLE property type accepting ALL sample types
    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='SAMPLE',
    )
    pt.save()

    assert pt.code == property_type_code.upper()
    assert pt.dataType == 'SAMPLE'
    assert pt.sampleType == '(ALL)'


def test_new_object_property_type_with_sample_type(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    sample_type_code = "test_sample_type_" + current_date + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    # Create custom SAMPLE property type accepting a single sample type
    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='SAMPLE',
        sampleType=sample_type.code,
    )
    pt.save()

    assert pt.code == property_type_code.upper()
    assert pt.dataType == 'SAMPLE'
    assert pt.sampleType == sample_type.code


def test_get_sample_type_from_property_assignment(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    sample_type_code = "test_sample_type_" + current_date + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='SAMPLE',
        sampleType=sample_type.code,
    )
    pt.save()

    sample_type.assign_property(pt)
    sample_type.save()


    ps = sample_type.get_property_assignments()

    assert len(ps.df) == 1

    property_type = ps[0].get_property_type()
    assert property_type.code.lower() == property_type_code
    assert property_type.dataType == 'SAMPLE'
    assert property_type.sampleType == sample_type.code


def test_get_sample_type_all_from_property_assignment(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    sample_type_code = "test_sample_type_" + current_date + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='SAMPLE',
    )
    pt.save()

    sample_type.assign_property(pt)
    sample_type.save()


    ps = sample_type.get_property_assignments()

    assert len(ps.df) == 1

    property_type = ps[0].get_property_type()
    assert property_type.code.lower() == property_type_code
    assert property_type.dataType == 'SAMPLE'
    assert property_type.sampleType == '(ALL)'


def test_get_vocabulary_code_from_property_assignment(space):
    o = space.openbis

    current_date = time.strftime("%Y-%m-%d").lower()

    voc_code = 'test_voc_' + current_date + "_" + str(uuid.uuid4())

    voc = o.new_vocabulary(
        code=voc_code,
        description='description of vocabulary',
        urlTemplate='https://ethz.ch',
        terms=[
            {"code": 'term_code1', "label": "term_label1", "description": "term_description1"},
            {"code": 'term_code2', "label": "term_label2", "description": "term_description2"},
            {"code": 'term_code3', "label": "term_label3", "description": "term_description3"}
        ],
        chosenFromList=False
    )
    voc.save()

    sample_type_code = "test_vocabulary_" + current_date + "_" + str(uuid.uuid4())
    sample_type = o.new_sample_type(
        code=sample_type_code,
        generatedCodePrefix="S",
        autoGeneratedCode=True,
        listable=True,
    )
    sample_type.save()

    property_type_code = "test_property_type_" + current_date + "_" + str(uuid.uuid4())
    pt = o.new_property_type(
        code=property_type_code,
        label='custom property of data type date',
        description='custom property created in unit test',
        dataType='CONTROLLEDVOCABULARY',
        vocabulary=voc_code
    )
    pt.save()

    sample_type.assign_property(pt)
    sample_type.save()


    ps = sample_type.get_property_assignments()

    assert len(ps.df) == 1

    property_type = ps[0].get_property_type()
    assert property_type.code.lower() == property_type_code
    assert property_type.dataType == 'CONTROLLEDVOCABULARY'
    assert property_type.vocabulary == voc_code.upper()


#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import binascii
import json
import os
import time
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

import pytest

from pybis.fast_download import FastDownload


def get_download_response(sequence_number, perm_id, file, is_directory, offset, payload):

    result = b''
    result += sequence_number.to_bytes(4, "big")
    download_item_id = perm_id + "/" + file
    result += len(download_item_id).to_bytes(2, "big")
    result += is_directory.to_bytes(1, "big")
    result += len(file).to_bytes(2, "big")
    # file offset here
    result += int(offset).to_bytes(8, "big")
    # payload length here
    payload_length = len(payload)
    result += int(payload_length).to_bytes(4, "big")
    result += download_item_id.encode()
    result += file.encode()
    result += binascii.crc32(result).to_bytes(8, "big")
    result += payload
    result += binascii.crc32(payload).to_bytes(8, "big")
    return result


class MyServer(BaseHTTPRequestHandler):
    """Message handler implementation for faking server. It is BaseHTTPRequestHandler is stateless
    so communication happens via static variable"""
    next_response = {}
    response_code = 200

    def do_POST(self):
        content_len = int(self.headers.get('Content-Length'))
        post_body = self.rfile.read(content_len)
        body = json.loads(post_body.decode("utf8"))

        method = body['method']
        if type(method) == list:
            method = method[0]
        if method == "download":
            response = next(MyServer.next_response['download'])
        else:
            response = MyServer.next_response[method].encode()
        self.send_response(MyServer.response_code)
        self.end_headers()
        self.wfile.write(response)


def create_fast_download_session(permId, files, download_url, wished_number_of_streams):
    return '''{ "jsonrpc": "2.0", "id": "2", "result": {
        "@type": "dss.dto.datasetfile.fastdownload.FastDownloadSession", "@id": 1,
        "downloadUrl": "''' + download_url + '''",
        "fileTransferUserSessionId": "admin-230317133255245x9E432E4241A58D972496ACD02AC95BC9",
        "files": [  { "@type": "dss.dto.datasetfile.id.DataSetFilePermId", "@id": 2,
                "dataSetId": { "@type": "as.dto.dataset.id.DataSetPermId", "@id": 3,
                    "permId": "''' + permId + '''"
                },  "filePath": "''' + permId + "/" + files + '''"  } ],
        "options": {
            "@type": "dss.dto.datasetfile.fastdownload.FastDownloadSessionOptions", "@id": 4,
            "wishedNumberOfStreams": ''' + wished_number_of_streams + ''' } } }'''


def start_download_session(ranges, wished_number_of_streams):
    return """{
        "downloadSessionId": "72863f8d-1ed1-4795-a531-4d93a5081562",
        "ranges": {
           """ + ", ".join([f'"{k}": "{v}"' for k, v in ranges.items()]) + """
        },
        "streamIds": [""" + ", ".join([f'"s{x}"' for x in range(int(wished_number_of_streams))]) \
        + """] }"""


def cleanup(directory):
    import shutil
    for files in os.listdir(directory):
        path = os.path.join(directory, files)
        try:
            shutil.rmtree(path)
        except OSError:
            os.remove(path)


@pytest.fixture(scope="session")
def base_data(tmp_path_factory):
    folder = tmp_path_factory.mktemp("temp_folder")
    port = 12354
    url = "127.0.0.1"
    download_url = "http://" + url + ":" + str(port)

    httpserver = HTTPServer((url, port), MyServer)
    Thread(target=httpserver.serve_forever, daemon=True).start()
    return folder, download_url


@pytest.fixture(autouse=True)
def run_around_tests(base_data):
    temp_folder = base_data[0]
    download_url = base_data[1]
    streams = "2"
    perm_id = "1"
    file = "original/test_file.txt"
    ranges = {
        "1/original/test_file.txt": "0:9"
    }
    MyServer.next_response = {
        'queue': "",
        'finishDownloadSession': "",
        'counter': 0,
        'parts': 10,
        'createFastDownloadSession': create_fast_download_session(perm_id,
                                                                  file,
                                                                  download_url,
                                                                  streams),
        'startDownloadSession': start_download_session(ranges, streams)
    }
    MyServer.response_code = 200
    yield temp_folder, download_url, streams, perm_id, file
    cleanup(temp_folder)

def get_server_information(major, minor):
    class ServerInfo:
        def is_version_greater_than(self, x, y):
            return x > major or (x == major and y > minor)

    return ServerInfo()

def test_download_fails_after_retries(run_around_tests):
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        while True:
            yield b''

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    try:
        fast_download.download()
        assert False
    except ValueError as error:
        assert str(error) == 'Consecutive download calls to the server failed.'


def test_download_file(run_around_tests):
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    fast_download.download()

    downloaded_files = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(temp_folder)
        for f in fn
    ]
    assert len(downloaded_files) == 1
    assert downloaded_files[0].endswith(file)
    import functools
    expected_outcome = functools.reduce(lambda a, b: a + b,
                                        [bytearray([x] * 10) for x in range(10)])
    with open(downloaded_files[0], 'rb') as fn:
        data = fn.read()
        assert len(data) == 100
        assert expected_outcome == data


def test_download_file_wait_flag_disabled(run_around_tests):
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            # Slow down responses to simulate download of a big file
            time.sleep(0.1)
            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, False, False, get_server_information(3,7), streams)
    fast_download.download()

    # Verify that file has not been downloaded yet
    downloaded_files = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(temp_folder)
        for f in fn
    ]
    assert len(downloaded_files) == 0

    # Wait for 2 seconds to finish download
    time.sleep(2)

    # find file
    downloaded_files = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(temp_folder)
        for f in fn
    ]
    assert len(downloaded_files) == 1

    assert downloaded_files[0].endswith(file)
    import functools
    expected_outcome = functools.reduce(lambda a, b: a + b,
                                        [bytearray([x] * 10) for x in range(10)])
    with open(downloaded_files[0], 'rb') as fn:
        data = fn.read()
        assert len(data) == 100
        assert expected_outcome == data


def test_download_file_starts_with_fail(run_around_tests):
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        MyServer.response_code = 502
        yield b''
        yield b''
        MyServer.response_code = 200
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    fast_download.download()

    downloaded_files = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(temp_folder)
        for f in fn
    ]
    assert len(downloaded_files) == 1
    assert downloaded_files[0].endswith(file)
    import functools
    expected_outcome = functools.reduce(lambda a, b: a + b,
                                        [bytearray([x] * 10) for x in range(10)])
    with open(downloaded_files[0], 'rb') as fn:
        data = fn.read()
        assert len(data) == 100
        assert expected_outcome == data


def test_download_fails_after_getting_java_exception(run_around_tests):
    """
    Test that verifies that if non-retryable exception is thrown,
    the whole download session is aborted.
    """
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        # First download fails with non-retryable exception
        yield b'{"error":"Some server error message.","retriable":false}'
        # Further responses are alright
        MyServer.response_code = 200
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    try:
        fast_download.download()
        assert False
    except ValueError as error:
        assert str(error) == 'Some server error message.'


def test_download_passes_after_getting_java_exception(run_around_tests):
    """
    Test that verifies that if retryable server exception is thrown,
    the whole download retries and downloads the file.
    """
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        # First download fails with non-retryable exception
        yield b'{"error":"Some server error message.","retriable":true}'
        # Further responses are alright
        MyServer.response_code = 200
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    fast_download.download()

    downloaded_files = [
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(temp_folder)
        for f in fn
    ]
    assert len(downloaded_files) == 1
    assert downloaded_files[0].endswith(file)
    import functools
    expected_outcome = functools.reduce(lambda a, b: a + b,
                                        [bytearray([x] * 10) for x in range(10)])
    with open(downloaded_files[0], 'rb') as fn:
        data = fn.read()
        assert len(data) == 100
        assert expected_outcome == data


def test_download_file_payload_failure(run_around_tests):
    temp_folder, download_url, streams, perm_id, file = run_around_tests

    def generate_download_response():
        parts = MyServer.next_response['parts']
        counter = MyServer.next_response['counter']
        payload_length = 10

        fail_response = None
        while counter < parts:
            response = get_download_response(counter, perm_id, file, False,
                                             counter * payload_length,
                                             bytearray([counter] * payload_length))
            if counter == 0:
                array = bytearray(response)
                array[-8:] = bytearray([0]*8)
                response = bytes(array)
                fail_response = response

            counter += 1
            MyServer.next_response['counter'] = counter % parts
            yield response

        while True:
            yield fail_response

    MyServer.next_response['download'] = generate_download_response()

    fast_download = FastDownload("", download_url, perm_id, file, str(temp_folder),
                                 True, True, False, get_server_information(3,7), streams)
    try:
        fast_download.download()
        assert False
    except ValueError as error:
        assert str(error) == 'Received incorrect payload multiple times. Aborting.'



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import random
import re

import pytest
import time
from random import randint
from pybis import DataSet
from pybis import Openbis


@pytest.fixture(scope='function')
def group(openbis_instance):
    group_name = 'test_group_{}'.format(randint(0, 1000)).upper()
    group = openbis_instance.new_group(
        code=group_name,
        description='description of group ' + group_name
    )
    group.save()
    yield group
    group.delete('test')


def test_crud_group(openbis_instance, group):
    group_exists = openbis_instance.get_group(group.code)
    assert group_exists is not None

    changed_description = 'changed description of group ' + group.code
    group.description = changed_description
    group.save()
    group_changed = openbis_instance.get_group(group.code)
    assert group_changed.description == changed_description

    group.delete('test')
    with pytest.raises(ValueError):
        group_not_exists = openbis_instance.get_group(group.code)
        assert group_not_exists is None


def test_group_member(openbis_instance, group):
    assert len(group.get_members()) == 0
    group.add_members('admin')
    group.save()
    assert len(group.get_members()) == 1

    group.del_members('admin')
    group.save()
    assert len(group.get_members()) == 0


def test_role_assignments(openbis_instance, group):
    roles_not_exist = group.get_roles()
    assert len(roles_not_exist) == 0

    group.assign_role('ADMIN')
    roles_exist = group.get_roles()
    assert len(roles_exist) == 1

    group.revoke_role('ADMIN')
    roles_not_exist = group.get_roles()
    assert len(roles_not_exist) == 0

    group.delete("test")

